// Sound volume ratio convertion functions
// (c) 2020 by Jin X (jin_x@list.ru)
// Delphi 6+

// This unit supports conversion between:
// * decibels,
// * amplitude and power ratio,
// * loudness ratio (in subjective perception).

unit SoundVolumeConverter;

{$IF CompilerVersion >= 17} // Delphi 2005+
  {$DEFINE INLINE}
{$IFEND}

interface


// Default constants
const
  // This constant can be used as second argument of *_DLdB functions (if you really need it).
  DefDblLoudnessDB           = 10.0;                     // default number of decibels for loudness doubling.
  // These two constants can be used as second argument of *_Exp functions (if you really need it).
  DefLoudnessToAmpRatioExp   = 1.660964047443681173935;  // default power exponent for conversion from loudness to amplitude ratio [Log2(DefDblLoudnessDB) / 2].
  DefLoudnessToPowerRatioExp = 3.321928094887362347870;  // default power exponent for conversion from loudness to power ratio [Log2(DefDblLoudnessDB)].
  // Next two constants is defined for internal use only (you shouldn't use them with any *_Exp functions, use constants defined above instead).
  DefAmpRatioToLoudnessExp   = 0.602059991327962390427;  // default power exponent for conversion from amplitude ratio to loudness [1 / DefLoudnessToAmpRatioExp].
  DefPowerRatioToLoudnessExp = 0.301029995663981195214;  // default power exponent for conversion from loudness to power ratio [1 / DefLoudnessToPowerRatioExp].


// Convert dB -> amplitude ratio.
function DecibelsToAmpRatio(const X: Extended): Extended; {$IFDEF INLINE}inline;{$ENDIF}
// Convert dB -> power ratio.
function DecibelsToPowerRatio(const X: Extended): Extended; {$IFDEF INLINE}inline;{$ENDIF}
// Convert dB -> amplitude/power ratio.
function DecibelsToValueRatio(const X: Extended; PowerRatio: Boolean): Extended;

// Convert dB -> loudness ratio.
function DecibelsToLoudnessRatio(const X: Extended): Extended; {$IFDEF INLINE}inline;{$ENDIF}
// Convert dB -> loudness ratio
// with user defined number of decibels for loudness doubling (DblLoudnessDB must be > 0; standard value is 10).
function DecibelsToLoudnessRatio_DLdB(const X, DblLoudnessDB: Extended): Extended; {$IFDEF INLINE}inline;{$ENDIF}
// Convert dB -> loudness ratio
// with user defined power exponent (LoudnessExp must be > 0; standard value is Log2(10)/2 for amplitude ratio and Log2(10) for power ratio).
// If PowerRatio = False (default) then LoudnessExp is specified as exponent for conversion to/from amplitude ratio, else to/from power ratio.
function DecibelsToLoudnessRatio_Exp(const X, LoudnessExp: Extended; PowerRatio: Boolean = False): Extended; {$IFDEF INLINE}inline;{$ENDIF}


// Convert amplitude ratio -> dB (X must be > 0).
function AmpRatioToDecibels(const X: Extended): Extended; {$IFDEF INLINE}inline;{$ENDIF}
// Convert power ratio -> dB (X must be > 0).
function PowerRatioToDecibels(const X: Extended): Extended; {$IFDEF INLINE}inline;{$ENDIF}
// Convert amplitude/power ratio -> dB (X must be > 0).
function ValueRatioToDecibels(const X: Extended; PowerRatio: Boolean): Extended;

// Convert amplitude ratio -> loudness ratio (X must be >= 0).
function AmpRatioToLoudnessRatio(const X: Extended): Extended; {$IFDEF INLINE}inline;{$ENDIF}
// Convert amplitude ratio -> loudness ratio (X must be >= 0).
function PowerRatioToLoudnessRatio(const X: Extended): Extended; {$IFDEF INLINE}inline;{$ENDIF}
// Convert amplitude/power ratio -> loudness ratio (X must be >= 0).
function ValueRatioToLoudnessRatio(const X: Extended; PowerRatio: Boolean): Extended;
// Convert amplitude/power ratio -> loudness ratio (X must be >= 0)
// with user defined power exponent (LoudnessExp must be > 0; standard value is Log2(10)/2 for amplitude ratio and Log2(10) for power ratio).
function ValueRatioToLoudnessRatio_Exp(const X, LoudnessExp: Extended): Extended; {$IFDEF INLINE}inline;{$ENDIF}

// Convert amplitude ratio -> loudness ratio (X must be >= 0)
// with user defined number of decibels for loudness doubling (DblLoudnessDB must be > 0; standard value is 10).
function AmpRatioToLoudnessRatio_DLdB(const X, DblLoudnessDB: Extended): Extended; {$IFDEF INLINE}inline;{$ENDIF}
// Convert power ratio -> loudness ratio (X must be >= 0)
// with user defined number of decibels for loudness doubling (DblLoudnessDB must be > 0; standard value is 10).
function PowerRatioToLoudnessRatio_DLdB(const X, DblLoudnessDB: Extended): Extended; {$IFDEF INLINE}inline;{$ENDIF}
// Convert amplitude/power ratio -> loudness ratio (X must be >= 0)
// with user defined number of decibels for loudness doubling (DblLoudnessDB must be > 0; standard value is 10).
function ValueRatioToLoudnessRatio_DLdB(const X, DblLoudnessDB: Extended; PowerRatio: Boolean): Extended; {$IFDEF INLINE}inline;{$ENDIF}


// Convert loudness ratio -> dB (X must be > 0).
function LoudnessRatioToDecibels(const X: Extended): Extended; {$IFDEF INLINE}inline;{$ENDIF}
// Convert loudness ratio -> dB (X must be > 0)
// with user defined number of decibels for loudness doubling (DblLoudnessDB must be > 0; standard value is 10).
function LoudnessRatioToDecibels_DLdB(const X, DblLoudnessDB: Extended): Extended; {$IFDEF INLINE}inline;{$ENDIF}
// Convert loudness ratio -> dB (X must be > 0)
// with user defined power exponent (LoudnessExp must be > 0; standard value is Log2(10)/2 for amplitude ratio and Log2(10) for power ratio).
// If PowerRatio = False (default) then LoudnessExp is specified as exponent for conversion to/from amplitude ratio, else to/from power ratio.
function LoudnessRatioToDecibels_Exp(const X, LoudnessExp: Extended; PowerRatio: Boolean = False): Extended; {$IFDEF INLINE}inline;{$ENDIF}

// Convert loudness ratio -> amplitude ratio (X must be >= 0).
function LoudnessRatioToAmpRatio(const X: Extended): Extended; {$IFDEF INLINE}inline;{$ENDIF}
// Convert loudness ratio -> power ratio (X must be >= 0).
function LoudnessRatioToPowerRatio(const X: Extended): Extended; {$IFDEF INLINE}inline;{$ENDIF}
// Convert loudness ratio -> amplitude/power ratio (X must be >= 0).
function LoudnessRatioToValueRatio(const X: Extended; PowerRatio: Boolean): Extended;
// Convert loudness ratio -> amplitude/power ratio (X must be >= 0)
// with user defined power exponent (LoudnessExp must be > 0; standard value is Log2(10)/2 for amplitude ratio and Log2(10) for power ratio).
function LoudnessRatioToValueRatio_Exp(const X, LoudnessExp: Extended): Extended; {$IFDEF INLINE}inline;{$ENDIF}

// Convert loudness ratio -> amplitude ratio (X must be >= 0)
// with user defined number of decibels for loudness doubling (DblLoudnessDB must be > 0; standard value is 10).
function LoudnessRatioToAmpRatio_DLdB(const X, DblLoudnessDB: Extended): Extended; {$IFDEF INLINE}inline;{$ENDIF}
// Convert loudness ratio -> power ratio (X must be >= 0)
// with user defined number of decibels for loudness doubling (DblLoudnessDB must be > 0; standard value is 10).
function LoudnessRatioToPowerRatio_DLdB(const X, DblLoudnessDB: Extended): Extended; {$IFDEF INLINE}inline;{$ENDIF}
// Convert loudness ratio -> amplitude/power ratio (X must be >= 0)
// with user defined number of decibels for loudness doubling (DblLoudnessDB must be > 0; standard value is 10).
function LoudnessRatioToValueRatio_DLdB(const X, DblLoudnessDB: Extended; PowerRatio: Boolean): Extended; {$IFDEF INLINE}inline;{$ENDIF}


// Convert <power exponent for conversion between loudness and amplitude/power ratio> to <number of decibels for loudness doubling> (LoudnessExp must be > 0).
// If PowerRatio = False (default) then LoudnessExp is specified as exponent for conversion to/from amplitude ratio, else to/from power ratio.
// Can be used as second argument of *_DLdB functions (if you really need it, by the way you can use *_Exp functions instead).
function LoudnessExpToDblLoudnessDB(LoudnessExp: Extended; PowerRatio: Boolean = False): Extended; {$IFDEF INLINE}inline;{$ENDIF}
// Convert <number of decibels for loudness doubling> to <power exponent for conversion between loudness and amplitude/power ratio> (DblLoudnessDB must be > 0).
// Can be used as second argument of *Exp functions (if you really need it, by the way you can use *_DLdB functions instead).
function DblLoudnessDBToLoudnessExp(const DblLoudnessDB: Extended; PowerRatio: Boolean = False): Extended; {$IFDEF INLINE}inline;{$ENDIF}


implementation

{$IF CompilerVersion >= 23} // Delphi XE2+
uses System.Math;
{$ELSE}
uses Math;
{$IFEND}

function DecibelsToAmpRatio(const X: Extended): Extended;
begin
  Result := Power(10, X / 20);
end;

function DecibelsToPowerRatio(const X: Extended): Extended;
begin
  Result := Power(10, X / 10);
end;

function DecibelsToValueRatio(const X: Extended; PowerRatio: Boolean): Extended;
begin
  if not PowerRatio then Result := Power(10, X / 20)  // amplitude ratio
  else Result := Power(10, X / 10);  // power ratio
end;

function DecibelsToLoudnessRatio(const X: Extended): Extended;
begin
  Result := Power(2, X / DefDblLoudnessDB);
end;

function DecibelsToLoudnessRatio_DLdB(const X, DblLoudnessDB: Extended): Extended;
begin
  Result := Power(2, X / DblLoudnessDB);
end;

function DecibelsToLoudnessRatio_Exp(const X, LoudnessExp: Extended; PowerRatio: Boolean = False): Extended;
begin
  Result := DecibelsToLoudnessRatio_DLdB(X, LoudnessExpToDblLoudnessDB(LoudnessExp, PowerRatio));
end;


function AmpRatioToDecibels(const X: Extended): Extended;
begin
  Result := Log10(X) * 20;
end;

function PowerRatioToDecibels(const X: Extended): Extended;
begin
  Result := Log10(X) * 10;
end;

function ValueRatioToDecibels(const X: Extended; PowerRatio: Boolean): Extended;
begin
  Result := Log10(X);
  if not PowerRatio then Result := Result * 20  // amplitude ratio
  else Result := Result * 10;  // power ratio
end;

function AmpRatioToLoudnessRatio(const X: Extended): Extended;
begin
  Result := Power(X, DefAmpRatioToLoudnessExp);
end;

function PowerRatioToLoudnessRatio(const X: Extended): Extended;
begin
  Result := Power(X, DefPowerRatioToLoudnessExp);
end;

function ValueRatioToLoudnessRatio(const X: Extended; PowerRatio: Boolean): Extended;
begin
  // alternative formulas for default exponent:
  // universal: Power(2, ValueRatioToDecibels(X, PowerRatio) / 10)
  // amplitude: Power(X, 2 / Log2(10))
  //     power: Power(X, 1 / Log2(10))
  if not PowerRatio then Result := Power(X, DefAmpRatioToLoudnessExp)  // amplitude ratio
  else Result := Power(X, DefPowerRatioToLoudnessExp);  // power ratio
end;

function ValueRatioToLoudnessRatio_Exp(const X, LoudnessExp: Extended): Extended;
begin
  Result := Power(X, 1 / LoudnessExp);  // power ratio
end;

function AmpRatioToLoudnessRatio_DLdB(const X, DblLoudnessDB: Extended): Extended;
begin
  Result := ValueRatioToLoudnessRatio_Exp(X, DblLoudnessDBToLoudnessExp(DblLoudnessDB, False));
end;

function PowerRatioToLoudnessRatio_DLdB(const X, DblLoudnessDB: Extended): Extended;
begin
  Result := ValueRatioToLoudnessRatio_Exp(X, DblLoudnessDBToLoudnessExp(DblLoudnessDB, True));
end;

function ValueRatioToLoudnessRatio_DLdB(const X, DblLoudnessDB: Extended; PowerRatio: Boolean): Extended;
begin
  Result := ValueRatioToLoudnessRatio_Exp(X, DblLoudnessDBToLoudnessExp(DblLoudnessDB, PowerRatio));
end;


function LoudnessRatioToDecibels(const X: Extended): Extended;
begin
  Result := Log2(X) * DefDblLoudnessDB;
end;

function LoudnessRatioToDecibels_DLdB(const X, DblLoudnessDB: Extended): Extended;
begin
  Result := Log2(X) * DblLoudnessDB;
end;

function LoudnessRatioToDecibels_Exp(const X, LoudnessExp: Extended; PowerRatio: Boolean = False): Extended;
begin
  Result := LoudnessRatioToDecibels_DLdB(X, LoudnessExpToDblLoudnessDB(LoudnessExp, PowerRatio));
end;

function LoudnessRatioToAmpRatio(const X: Extended): Extended;
begin
  Result := Power(X, DefLoudnessToAmpRatioExp);
end;

function LoudnessRatioToPowerRatio(const X: Extended): Extended;
begin
  Result := Power(X, DefLoudnessToPowerRatioExp);
end;

function LoudnessRatioToValueRatio(const X: Extended; PowerRatio: Boolean): Extended;
begin
  // alternative formulas for default exponent:
  // amplitude: Power(X, Log2(10) / 2) ; Power(10, LoudnessRatioToDecibels(X) / 20)
  //     power: Power(X, Log2(10)) ; Power(10, LoudnessRatioToDecibels(X) / 10)
  if not PowerRatio then Result := Power(X, DefLoudnessToAmpRatioExp)  // amplitude ratio
  else Result := Power(X, DefLoudnessToPowerRatioExp);  // power ratio
end;

function LoudnessRatioToValueRatio_Exp(const X, LoudnessExp: Extended): Extended;
begin
  Result := Power(X, LoudnessExp);
end;

function LoudnessRatioToAmpRatio_DLdB(const X, DblLoudnessDB: Extended): Extended;
begin
  Result := LoudnessRatioToValueRatio_Exp(X, DblLoudnessDBToLoudnessExp(DblLoudnessDB, False));
end;

function LoudnessRatioToPowerRatio_DLdB(const X, DblLoudnessDB: Extended): Extended;
begin
  Result := LoudnessRatioToValueRatio_Exp(X, DblLoudnessDBToLoudnessExp(DblLoudnessDB, True));
end;

function LoudnessRatioToValueRatio_DLdB(const X, DblLoudnessDB: Extended; PowerRatio: Boolean): Extended;
begin
  Result := LoudnessRatioToValueRatio_Exp(X, DblLoudnessDBToLoudnessExp(DblLoudnessDB, PowerRatio));
end;


function LoudnessExpToDblLoudnessDB(LoudnessExp: Extended; PowerRatio: Boolean = False): Extended;
begin
  if not PowerRatio then LoudnessExp := LoudnessExp * 2;  // amplitude ratio
  Result := Power(2, LoudnessExp);
end;

function DblLoudnessDBToLoudnessExp(const DblLoudnessDB: Extended; PowerRatio: Boolean = False): Extended;
begin
  Result := Log2(DblLoudnessDB);
  if not PowerRatio then Result := Result * 0.5;  // amplitude ratio
end;


end.
