program SoundVolumeRatioConverter;
{$APPTYPE CONSOLE}
// Delphi 7

uses SysUtils, Math, SoundVolumeConverter;

var
  Mode: (Unknown, dB, Ratio, Loudness) = Unknown;
  S: String;
  N: String = '';
  Param: Integer = 0;
  LoudnessExp: Extended = DefLoudnessToAmpRatioExp;
  LoudnessExpType: (deAmp, dePower, deAuto) = deAmp;
  PowerValues: Boolean = False;
  ValueText: String = 'amplitude';
  Percent: Boolean = False;
  Precision: Integer = -1;
  TrailingZeroes: Boolean = False;
  ShowLogo: Boolean = True;
  Quiet: Boolean = False;
  ShowSource: Boolean = False;       // only when Quiet
  ShowDB: Boolean = False;           // only when Quiet
  ShowRatio: Boolean = False;        // only when Quiet
  ShowLoudness: Boolean = False;     // only when Quiet
  RatioReciprocal: Boolean = False;  // only when Quiet
  LoudnessPercent: Boolean = False;  // only when Quiet
  ShowUnits: Boolean = True;         // only when Quiet
  Pause: Boolean = False;
  Help: Boolean = False;

// Get next command line parameter
function NextParamStr: String;
begin
  Inc(Param);
  Result := UpperCase(Trim(ParamStr(Param)));
end;

// Convert float to string depending on settings
function FloatToString(N: Extended): String;
begin
  if Precision = -1 then Result := FloatToStr(N)
  else if TrailingZeroes then Result := FloatToStrF(N, ffFixed, 16, Precision)
  else Result := FloatToStr(RoundTo(N, -Precision));
end;

// Error message with halt
procedure Error(const S: String);
begin
  WriteLn(S, ', aborting...');
  Halt(1);
end;

// Convert string to float with error checking
function GetFloat(const S: String; const ErrMsg: String = 'Wrong value'): Extended;
begin
  if not TryStrToFloat(S, Result) then Error(ErrMsg);
end;

// Convert string with number or ratio to float
function GetValue(S: String; Frac: Boolean = False): Extended;
var k: Integer;
begin
  S := StringReplace(S, ',', '.', [rfReplaceAll]);
  if Frac then k := Pos('/', S)
  else k := 0;
  if k = 0 then Result := GetFloat(S)
  else
  begin
    Result := GetFloat(Copy(S, k+1, MaxInt));
    if Result = 0 then Error('Division by zero');
    Result := GetFloat(Copy(S, 1, k-1)) / Result;
  end;
  if Percent then Result := Result / 100;
end;

// Process command line parameters
procedure ProcessParams;
var
  i: Integer;
  Ch: Char;
begin
  repeat
    S := NextParamStr;
    if S = '' then Break;
    if (S = '-?') or (S = '/?') or (S = '-HELP') or (S = '/HELP') or (S = '--HELP') then Help := True
    else if (S = '-D') or (S = '/D') then Mode := dB
    else if (S = '-R') or (S = '/R') then Mode := Ratio
    else if (S = '-S') or (S = '/S') then Mode := Loudness
    else if (S = '-P') or (S = '/P') then
    begin
      PowerValues := True;
      ValueText := 'power';
    end
    else if (Copy(S, 1, 2) = '-X') or (Copy(S, 1, 2) = '/X') then
    begin
      Delete(S, 1, 2);
      Ch := #0;
      if (S <> '') and (S[1] in ['A'..'Z']) then
      begin
        Ch := S[1];
        Delete(S, 1, 1);
      end;
      LoudnessExp := GetFloat(S, 'Wrong value for -X option');
      case Ch of
        #0: LoudnessExpType := deAuto;
        'A': LoudnessExpType := deAmp;
        'P': LoudnessExpType := dePower;
        'D':
        begin
          LoudnessExp := DblLoudnessDBToLoudnessExp(LoudnessExp, True);
          LoudnessExpType := dePower;
        end
        else Error('Wrong -X suboption');
      end;
    end
    else if (Copy(S, 1, 2) = '-F') or (Copy(S, 1, 2) = '/F') then
    begin
      Delete(S, 1, 2);
      if (S <> '') and (S[1] = 'Z') then
      begin
        TrailingZeroes := True;
        Delete(S, 1, 1);
      end;
      if S = '' then Precision := 6
      else if not TryStrToInt(S, Precision) or (Precision < 0) or (Precision > 37) then Error('Wrong -F option value');
    end
    else if (Copy(S, 1, 2) = '-Q') or (Copy(S, 1, 2) = '/Q') then
    begin
      Quiet := True;
      Delete(S, 1, 2);
      for i := 1 to Length(S) do
        case S[i] of
          'S': ShowSource := True;
          'D': ShowDB := True;
          'R': ShowRatio := True;
          'L': ShowLoudness := True;
          'I': RatioReciprocal := True;
          'P': LoudnessPercent := True;
          'U': ShowUnits := False
          else Error('Wrong -Q suboption');
        end;
    end
    else if (S = '-NL') or (S = '/NL') then ShowLogo := False
    else if (S = '-W') or (S = '/W') then Pause := True
    else if (N = '') and (S[1] <> '/') then N := S
    else Error('Wrong parameters');
  until False;
end;

// Show title or help
procedure TitleAndHelp;
begin
  if not Quiet and ShowLogo then
  begin
    if Help then WriteLn('--------------------------------------------------------------------------');
                 WriteLn('Sound Volume Ratio Converter, v1.36 (c) 2017-2020 by Jin X (jin_x@list.ru)');
    if Help then WriteLn('--------------------------------------------------------------------------');
  end;
  if Help then
  begin
    if not Quiet then
    begin
      if ShowLogo then WriteLn;
      WriteLn('Usage:');
      Write('  ');
    end;
    WriteLn('SVRC [X[/Y][x|%]] [-D|-R|-L] [-P] [-X[A|P]n] [-F[Z][n]] [-Q[S][D][R][L][I][P]] [-NL] [-W] [-?|--HELP]');
    if not Quiet then
    begin
      WriteLn;
      WriteLn('where:');
      WriteLn('  X or X/Y - number (X) or fraction (X/Y), fraction can be specified only for conversion from amplitude ratio!');
      WriteLn('  Xx, X/Yx, X% or X/Y% - loudness ratio (in subjective perception).');
      WriteLn('       (where ''x'' is latin character), Xx = 100*X%.');
      WriteLn('  -D - convert decibels -> amplitude ratio and loudness ratio [default if just number X is specified].');
      WriteLn('  -R - convert amplitude ratio -> decibels and loudness ratio [default if just fraction X/Y is specified].');
      WriteLn('  -L - convert loudness ratio -> decibels and amplitude ratio [default if ''x'' or ''%'' character is added to X or X/Y].');
      WriteLn('  -P - use power (energy values) ratio instead of amplitude/voltage/acoustic pressure/distance (force/field values);');
      WriteLn('       power (energy) values are proportional to square of amplitude (force) values.');
      WriteLn('  -X[A|P|D]n - set power exponent for loudness ratio (A - for amplitude ratio conversion, P - for power ratio conversion, D - value in decibels');
      WriteLn('       (see below), else - depending on -P option presence); when -XA is specified then -XP value is calculated as double of -XA (-XA as a');
      WriteLn('       half of -XP); e.g. -XA2 equals to -XP4, -XA3 equals to -XP6, -XA1.66 equals to -XP3.22, so you can''t specify two independent values');
      WriteLn('       for -XA and -XP options; suboption D points that value is specified in decibels per 2x (200%) loudness ratio instead of power exponent:');
      WriteLn('       n value for -XD option equals to log2(n)/2 value for -XA and to log2(n) value for -XP [default value is 1.660964047443681173935 for');
      WriteLn('       amplitude ratio (-XA) = 3.32192809488736234787 for power ratio (-XP) = 10 dB (-XD), this is standard value for human ears].');
      WriteLn('  -F[Z][n] - set precision of displaying real number fractional part to n digits [default is 6]; Z - display trailing zeros.');
      WriteLn('  -Q[S][D][R][L][I][P][U] - don''t show extra messages; S - show source value, D - show destination decibels, R - show destination amplitude ratio,');
      WriteLn('       L - show destination loudness ratio; I - show reciprocal value of amplitude ratio, P - show loudness in percents (options I, P are relevant');
      WriteLn('       only if displaying of corresponding values is enabled); U - don''t show units [if no D, R or L suboptions are specified then R is used when');
      WriteLn('       converting from decibels, D in other cases].');
      WriteLn('  -NL - don''t show logo and copyright string.');
      WriteLn('  -W - pause (wait for Enter key) after conversion.');
      WriteLn('  -?|--HELP - show this help (can be combined with other options: -P, -Q, -W).');
      WriteLn;
      WriteLn('Notes:');
      WriteLn('* Programs sets ERRORLEVEL = 1 in case of error.');
      WriteLn('* All options are case-insensitive and can be started with slash (/) instead of hyphen (-).');
      WriteLn;
      WriteLn('Examples:');
      WriteLn('  SVRC 10/3 - convert 10/3 amplitude ratio to decibels and loudness ratio.');
      WriteLn('  SVRC -6 /D /Q - convert -6 dB to amplitude ratio with no extra messages.');
      WriteLn('  SVRC 50% -S -W - convert 50% (the same as 1/2x) of loudness ratio to decibels and amplitude ratio with pause after conversion.');
      WriteLn('  SVRC 1/5x -P -QSPRI - convert 1/5x (20%) of loudness ratio to reciprocal of power ratio with displaying of source value in percents.');
      WriteLn;
      if PowerValues then WriteLn('Formulas:')
      else
        WriteLn('Formulas (use -P -? combination to see formulas for power (energy) ratios):');
      if PowerValues then
        WriteLn('  dB -> power ratio                 : 10 ^ (X/10).')
      else
        WriteLn('  dB -> amplitude ratio             : 10 ^ (X/20).');
        WriteLn('  dB -> loudness ratio              : 2 ^ (X/10).');
      if PowerValues then
      begin
        WriteLn('  power ratio -> dB                 : 10 * log10(X).');
        WriteLn('  power ratio -> loudness ratio     : X ^ (1/log2(10)) = X ^ 0.301029995663981195214 [power exponent is reciprocal of -XP option value].');
      end
      else
      begin
        WriteLn('  amplitude ratio -> dB             : 20 * log10(X).');
        WriteLn('  amplitude ratio -> loudness ratio : X ^ (2/log2(10)) = X ^ 0.602059991327962390427 [power exponent is reciprocal of -XA option value].');
      end;
        WriteLn('  loudness ratio -> dB              : 10 * log2(X).');
      if PowerValues then
        WriteLn('  loudness ratio -> power ratio     : X ^ log2(10) = X ^ 3.32192809488736234787 [power exponent is -XP option value].')
      else
        WriteLn('  loudness ratio -> amplitude ratio : X ^ (log2(10)/2) = X ^ 1.660964047443681173935 [power exponent is -XA option value].');
    end;
    if Pause then ReadLn;
    Halt;
  end;
end;

// Enter conversion mode if necessary
procedure EnterMode;
begin
  if (Mode = Unknown) and (N = '') then
  begin
    if not Quiet then
    begin
      if ShowLogo then
      begin
        WriteLn('Specify -? in command line to get parameters info.');
        WriteLn;
      end;
      WriteLn('Enter ''D'' or just number X to convert  : decibels -> ', ValueText, ' ratio and loudness ratio.');
      WriteLn('Enter ''R'' or fraction X/Y to convert   : ', ValueText, ' ratio -> decibels and loudness ratio.');
      WriteLn('Enter ''L'' or Xx, X% or X/Yx to convert : loudness ratio -> decibels and ', ValueText, ' ratio.');
      Write('> ');
    end;
    ReadLn(S);
    S := UpperCase(Trim(S));
    if S = 'D' then Mode := dB
    else if S = 'R' then Mode := Ratio
    else if S = 'L' then Mode := Loudness
    else
    begin
      N := S;
      if N = '' then N := ' ';
    end;
  end;
  if not Quiet and ShowLogo then WriteLn;
end;

// Correct parameters (Mode, Quiet, LoudnessExp)
procedure CorrectParams;
begin
  if Mode = Unknown then
  begin
    if (Pos('%', N) = Length(N)) or (Pos('X', N) = Length(N)) then Mode := Loudness
    else if Pos('/', N) > 0 then Mode := Ratio
    else Mode := dB;
  end;

  if Quiet and (not ShowDB or (Mode = dB)) and (not ShowRatio or (Mode = Ratio)) and (not ShowLoudness or (Mode = Loudness)) then
  begin
    if Mode = dB then ShowRatio := True
    else ShowDB := True;
  end;

  if (LoudnessExpType = deAmp) and PowerValues then LoudnessExp := LoudnessExp * 2
  else if (LoudnessExpType = dePower) and not PowerValues then LoudnessExp := LoudnessExp / 2;
end;

// Enter source value (is necessary)
procedure CheckAndEnterX(const Prompt: String);
begin
  if N = '' then
  begin
    if not Quiet then Write(Prompt);
    ReadLn(N);
    N := UpperCase(Trim(N));
  end;
end;

// Enter source value (is necessary), convert from decibels and display results
procedure ConvertFromDB;
var X, Y, Z: Extended;
begin
  CheckAndEnterX('Enter dB value: ');
  X := GetValue(N);

  Y := DecibelsToValueRatio(X, PowerValues);
  Z := DecibelsToLoudnessRatio_Exp(X, LoudnessExp, PowerValues);

  if not Quiet or ShowSource or ShowDB then
  begin
    Write(FloatToString(X));
    if Quiet then
    begin
      if ShowUnits then Write(' dB');
      Write(' : ');
    end;
  end;
  if not Quiet then
  begin
    WriteLn(' dB.');
    WriteLn(FloatToString(Y), '/1 (1/', FloatToString(1/Y), ') ', ValueText, ' ratio.');
    Write(FloatToString(Z), 'x (', FloatToString(100*Z), '%) loudness ratio (in subjective perception).');
  end
  else
  begin
    if ShowRatio then
    begin
      if RatioReciprocal then Y := 1 / Y;
      Write(FloatToString(Y));
      if ShowUnits then Write('/1');
      if ShowLoudness then Write(' : ');
    end;
    if ShowLoudness then
    begin
      if LoudnessPercent then Z := Z * 100;
      Write(FloatToString(Z));
      if ShowUnits then
        if LoudnessPercent then Write('%') else Write('x');
    end;
  end;
  WriteLn;
end;

// Enter source value (is necessary), convert from ratio and display results
procedure ConvertFromRatio;
var X, Y, Z: Extended;
begin
  CheckAndEnterX('Enter ' + ValueText + ' ratio (can be just value or fraction X/Y): ');
  X := GetValue(N, True);
  if X <= 0 then Error('Value of ' + ValueText + ' ratio must be positive');

  Y := ValueRatioToDecibels(X, PowerValues);
  Z := ValueRatioToLoudnessRatio_Exp(X, LoudnessExp);

  if not Quiet or ShowSource or ShowRatio then
  begin
    if RatioReciprocal then X := 1 / X;
    Write(FloatToString(X));
    if Quiet then
    begin
      if ShowUnits then Write('/1');
      Write(' : ');
    end;
  end;
  if not Quiet then
  begin
    WriteLn(' (1/', FloatToString(1/X), ') ', ValueText, ' ratio.');
    WriteLn(FloatToString(Y), ' dB.');
    Write(FloatToString(Z), 'x (', FloatToString(100*Z), '%) loudness ratio (in subjective perception).');
  end
  else
  begin
    if ShowDB then
    begin
      Write(FloatToString(Y));
      if ShowUnits then Write(' dB');
      if ShowLoudness then Write(' : ');
    end;
    if ShowLoudness then
    begin
      if LoudnessPercent then Z := Z * 100;
      Write(FloatToString(Z));
      if ShowUnits then
        if LoudnessPercent then Write('%') else Write('x');
    end;
  end;
  WriteLn;
end;

// Enter source value (is necessary), convert from loudness ratio and display results
procedure ConvertFromLoudness;
var X, Y, Z: Extended;
begin
  CheckAndEnterX('Enter perceived loudness ratio (can be just value, percent X% or fraction X/Y): ');
  if Pos('%', N) = Length(N) then Percent := True;
  if Percent or (Pos('X', N) = Length(N)) then Delete(N, Length(N), 1);

  X := GetValue(N, True);
  if X <= 0 then Error('Loudness ratio value must be positive');

  Y := LoudnessRatioToDecibels_Exp(X, LoudnessExp, PowerValues);
  Z := LoudnessRatioToValueRatio_Exp(X, LoudnessExp);

  if not Quiet then
  begin
    WriteLn(FloatToString(X), 'x (', FloatToString(100*X), '%) loudness ratio (in subjective perception).');
    WriteLn(FloatToString(Y), ' dB.');
    Write(FloatToString(Z), ' (1/', FloatToString(1/Z), ') ', ValueText, ' ratio.');
  end
  else
  begin
    if ShowSource or ShowLoudness then
    begin
      if LoudnessPercent then X := X * 100;
      Write(FloatToString(X));
      if ShowUnits then
        if LoudnessPercent then Write('%') else Write('x');
      Write(' : ');
    end;
    if ShowDB then
    begin
      Write(FloatToString(Y));
      if ShowUnits then Write(' dB');
      if ShowRatio then Write(' : ');
    end;
    if ShowRatio then
    begin
      if RatioReciprocal then Z := 1 / Z;
      Write(FloatToString(Z));
      if ShowUnits then Write('/1');
    end;
  end;
  WriteLn;
end;

// Main procedure
begin
  DecimalSeparator := '.';
  // Process command line parameters
  ProcessParams;
  TitleAndHelp;
  // Enter mode (if necessary) and correct parameters
  EnterMode;
  CorrectParams;
  // Enter source value (if necessary), convert and display results
  if Mode = dB then ConvertFromDB
  else if Mode = Ratio then ConvertFromRatio
  else ConvertFromLoudness;
  // Wait for Enter keypress
  if Pause then ReadLn;
end.
